import logging
from typing import Mapping, Any, List, Iterator, Tuple


from gladier.exc import FlowGenException

log = logging.getLogger(__name__)

STATE_TRANSITION_KEYS = {"Next", "Default", "StartAt"}


def get_transition(state):
    for k in STATE_TRANSITION_KEYS:
        if k in state:
            return k, state[k]


def iter_flow_states(
    flow_states: Mapping[str, Any],
    state: str,
    previously_visited: List[str] = None,
) -> Iterator[Tuple[str, Mapping[str, Any]]]:
    """Iter a Depth first search through a given dict of flow states. Previously
    visited states are skipped."""
    visited = previously_visited.copy() if previously_visited else list()

    if state is None or state in visited:
        return
    visited.append(state)

    if state not in flow_states:
        raise FlowGenException(f'State {state} not in definition!')

    yield state, flow_states[state]

    transition_state = get_transition(flow_states[state])
    if transition_state:
        _, state_name = transition_state
        yield from iter_flow_states(flow_states, state_name, previously_visited=visited)

    state_info = flow_states[state]
    if state_info['Type'] == 'Choice':
        for choice in state_info.get('Choices', []):
            if choice.get('Next'):
                yield from iter_flow_states(flow_states, choice['Next'], previously_visited=visited)
    return


def iter_flow(flow_definition: Mapping[str, Any]) -> Iterator[Tuple[str, Mapping[str, Any]]]:
    """
    Yields a tuple containing the state name and state info for each state in a flow definition.

    """
    yield from iter_flow_states(flow_definition['States'], flow_definition['StartAt'])


def get_end_states(flow_definition: Mapping[str, Any]) -> Iterator[str]:
    """Get all states for a flow that will cause the flow to exit normally. This includes any
    state which contains "End": True. A termination state is assumed if there is no next
    state to transition."""
    yield from {name for name, state in iter_flow(flow_definition) if not get_transition(state)}



from __future__ import annotations

import logging
import globus_sdk
from globus_sdk.scopes import AuthScopes, SearchScopes
from funcx.sdk.web_client import FuncxWebClient
from funcx import FuncXClient

log = logging.getLogger(__name__)


class FuncXLoginManager:
    """
    Implements the funcx.sdk.login_manager.protocol.LoginManagerProtocol class.

    https://github.com/funcx-faas/funcX/blob/main/funcx_sdk/funcx/sdk/login_manager/protocol.py#L18
    """
    SCOPES = [
        FuncXClient.FUNCX_SCOPE,
        AuthScopes.openid,
        SearchScopes.all
    ]

    def __init__(self, authorizers: dict[str, globus_sdk.RefreshTokenAuthorizer]):
        self.authorizers = authorizers

    def get_auth_client(self) -> globus_sdk.AuthClient:
        return globus_sdk.AuthClient(
            authorizer=self.authorizers[AuthScopes.openid]
        )

    def get_search_client(self) -> globus_sdk.SearchClient:
        return globus_sdk.SearchClient(
            authorizer=self.authorizers[SearchScopes.all]
        )

    def get_funcx_web_client(
        self, *, base_url: str | None = None, app_name: str | None = None
    ) -> FuncxWebClient:
        return FuncxWebClient(
            base_url=base_url,
            app_name=app_name,
            authorizer=self.authorizers[FuncXClient.FUNCX_SCOPE],
        )

    def ensure_logged_in(self):
        log.warning("ensure_logged_in cannot be invoked from here!")

    def logout(self):
        log.warning("logout cannot be invoked from here!")



import os
import stat
from fair_research_login.token_storage import flat_pack, flat_unpack
from gladier.storage.config import GladierConfig


class GladierSecretsConfig(GladierConfig):

    DEFAULT_PERMISSION = stat.S_IRUSR | stat.S_IWUSR

    def save(self):
        super().save()
        os.chmod(self.filename, self.DEFAULT_PERMISSION)

    def write_tokens(self, tokens):
        self.load()
        for name, value in flat_pack(tokens).items():
            self.set(self.section, name, value)
        self.save()

    def read_tokens(self):
        self.load()
        return flat_unpack(dict(self.items(self.section)))

    def clear_tokens(self):
        self.load()
        self.remove_section(self.section)
        self.add_section(self.section)
        self.save()



import abc
import copy
import logging
import gladier.utils.name_generation

log = logging.getLogger(__name__)


class ToolAlias(abc.ABC):

    input_location = 'input'

    def __init__(self, alias):
        self.alias = alias

    @abc.abstractmethod
    def rename_state(self, state_name, tool):
        return state_name

    @abc.abstractmethod
    def rename_variable(self, variable_name, tool):
        return variable_name

    def get_input_variable(self, flow_input_variable, tool_inputs):
        location = f'$.{self.input_location}.'
        input_name = flow_input_variable.replace(location, '')
        if input_name in tool_inputs:
            return input_name

    def rename_input_variables(self, state_data, tool_inputs, tool):
        sdata = copy.deepcopy(state_data)
        if not sdata:
            return
        for k in sdata.keys():
            if isinstance(sdata[k], str):
                input_var = self.get_input_variable(sdata[k], tool_inputs)
                if input_var:
                    new_var = f'$.{self.input_location}.{self.rename_variable(input_var, tool)}'
                    sdata[k] = new_var
            elif isinstance(sdata[k], dict):
                sdata[k] = self.rename_input_variables(sdata[k], tool_inputs, tool)
            elif isinstance(sdata[k], list):
                sdata[k] = [self.rename_input_variables(subsdata, tool_inputs, tool)
                            for subsdata in sdata[k]]
        return sdata


class NoAlias(ToolAlias):
    def rename_state(self, state_name, tool):
        return state_name

    def rename_variable(self, variable_name, tool):
        return variable_name


class StateSuffixVariablePrefix(ToolAlias):

    def rename_state(self, state_name, tool):
        return f'{state_name}{self.alias}'

    def rename_variable(self, variable_name, tool):
        return f'{gladier.utils.name_generation.get_snake_case(self.alias)}_{variable_name}'




import logging
from typing import List
import abc
from gladier.managers.login_manager import BaseLoginManager
from gladier.storage.config import GladierConfig


log = logging.getLogger(__name__)


class ServiceManager(abc.ABC):

    def __init__(self, **kwargs):
        self._storage = kwargs.get('storage')

        self.login_manager = kwargs.get('login_manager')
        self.register_scopes()

    @property
    def storage(self):
        if self._storage is None:
            raise AttributeError(f'Storage has not been set for {self}')
        return self._storage

    @storage.setter
    def storage(self, value):
        self.set_storage(value)

    def set_storage(self, storage: GladierConfig, replace: bool = True) -> None:
        if replace:
            log.info(f'Replacing storage {self._storage} with {storage}')
            self._storage = storage
        else:
            self._storage = self._storage or storage
        log.debug(f'Storage for {self} set to {self._storage}')

    def set_login_manager(self, login_manager: BaseLoginManager, replace: bool = True) -> None:
        if replace:
            log.info(f'Replacing login manager {self.login_manager} with {login_manager}')
            self.login_manager = login_manager
        else:
            self.login_manager = self.login_manager or login_manager
        self.login_manager.add_requirements(self.get_scopes())
        log.debug(f'Login Manager for {self} set to {self.login_manager}')

    def register_scopes(self):
        """
        Register scopes returned by `get_scopes()` with the login manager"""
        if self.login_manager is not None:
            self.login_manager.add_requirements(self.get_scopes())

    def get_scopes() -> List[str]:
        raise NotImplementedError('Service Managers must define their own scope needs')



import os
import logging
import warnings
from typing import Union, Mapping
from collections.abc import Iterable

from gladier.base import GladierBaseTool
from gladier.managers.login_manager import (
    BaseLoginManager, CallbackLoginManager, AutoLoginManager
)
from globus_sdk import AccessTokenAuthorizer, RefreshTokenAuthorizer
from gladier.managers import FlowsManager, FuncXManager

from gladier.storage.tokens import GladierSecretsConfig
from gladier.storage.config import GladierConfig
import gladier
import gladier.storage.config
import gladier.utils.dynamic_imports
import gladier.utils.automate
import gladier.utils.name_generation
import gladier.storage.migrations
import gladier.utils.tool_alias
import gladier.utils.funcx_login_manager
import gladier.exc
import gladier.version
log = logging.getLogger(__name__)


class GladierBaseClient(object):
    """
    The Gladier Client ties together commonly used funcx functions
    and basic flows with auto-registration tools to make complex tasks
    easy to automate.

    This class is intended to be subclassed as follows:

    .. code-block:: python

        @generate_flow_definition
        class MyGladierClient(GladierBaseClient):
            gladier_tools = [MyTool]

    And used like the following:

    .. code-block:: python

        my_gc = MyGladierClient()
        flow = my_gc.run_flow(flow_input={"flow_input": {"my_field": "foo"}})
        run_id = flow["run_id"]
        my_gc.progress(run_id)
        pprint(tar_and_transfer.get_status(run_id))

    The following class variables can be set on clients to change their behavior when
    deploying and running flows.

    * glaider_tools (default: [])
       * A list of Gladier Tools to build a working flow_defitinion. Each tool's minimum input
         must be satisfied prior to running the flow. Can be used with the
         @generate_flow_definition decorator to automatically chain together flow definitions
         present on each tool in linear order.
    * flow_definition (default: {})
       * An explicit flow definition to use for this client. Cannot be used with
         @generate_flow_definition
    * secret_config_filename (default: ``~/.gladier-secrets.cfg``)
       * Storage are for Globus Tokens and general storage
    * app_name (default: 'Gladier Client')
       * The app name used during a login flow
    * client_id
       * The Globus Client ID used for native logins
    * globus_group (default: None)
       * A Globus Group to be applied to all flow/run permissions. Group will automatically be
         added to flow_viewers, flow_starters, flow_administrators, run_managers, run_monitors
    * subscription_id (default: None)
       * The subscription id associated with this flow
    * alias_class (default: gladier.utils.tool_alias.StateSuffixVariablePrefix)
       * The default class used to for applying aliases to Tools

    Default options are intended for CLI usage and maximum user convenience.

    :param auto_registration: Automatically register functions or flows if they are not
                              previously registered or obsolete.
    :param login_manager: Class defining login behavior. Defaults to AutoLoginManager, and
                          will auto-login when additional scopes are needed.
    :param flows_manager: A flows manager class with customized behavior. Attrs like group
                          and login_manager will automatically be set if None
    :raises gladier.exc.AuthException: if authorizers given are insufficient

    """
    secret_config_filename = os.path.expanduser("~/.gladier-secrets.cfg")
    app_name = 'Gladier Client'
    client_id = 'f1631610-d9e4-4db2-81ba-7f93ad4414e3'
    globus_group = None
    subscription_id = None
    alias_class = gladier.utils.tool_alias.StateSuffixVariablePrefix

    def __init__(
        self,
        authorizers: Mapping[str, Union[AccessTokenAuthorizer, RefreshTokenAuthorizer]] = None,
        auto_login: bool = True,
        auto_registration: bool = True,
        login_manager: BaseLoginManager = None,
        flows_manager: FlowsManager = None,
            ):

        self._tools = None

        # Setup storage
        section = gladier.utils.name_generation.get_snake_case(self.__class__.__name__)
        self.storage = GladierConfig(self.secret_config_filename, section)
        self.storage.update()

        if auto_login is False:
            warnings.warn('auto_login=False in Gladier clients is deprecated and will '
                          'be removed in v0.8. See '
                          'https://gladier.readthedocs.io/en/latest/gladier/customizing_auth.html',
                          category=DeprecationWarning)
        if authorizers:
            warnings.warn('Calling Gladier clients with "authorizers" is deprecated. Instead, see '
                          'https://gladier.readthedocs.io/en/latest/gladier/customizing_auth.html',
                          category=DeprecationWarning)
            self.login_manager = CallbackLoginManager(authorizers=authorizers)
        elif not login_manager:
            section_name = f'tokens_{self.client_id}'
            token_storage = GladierSecretsConfig(self.secret_config_filename, section_name)
            self.login_manager = AutoLoginManager(self.client_id, token_storage, self.app_name,
                                                  auto_login=auto_login)
        else:
            self.login_manager = login_manager

        self.flows_manager = flows_manager or FlowsManager(auto_registration=auto_registration)
        if self.globus_group:
            self.flows_manager.globus_group = self.globus_group
        if self.subscription_id:
            self.flows_manager.subscription_id = self.subscription_id
        if not self.flows_manager.flow_title:
            self.flows_manager.flow_title = f'{self.__class__.__name__} flow'

        self.funcx_manager = FuncXManager(auto_registration=auto_registration)

        for man in (self.flows_manager, self.funcx_manager):
            man.set_storage(self.storage, replace=False)
            man.set_login_manager(self.login_manager, replace=False)
            man.register_scopes()

    @staticmethod
    def get_gladier_defaults_cls(tool_ref, alias_class=None):
        """
        Load a Gladier default class (gladier.GladierBaseTool) by import string. For
        Example: get_gladier_defaults_cls('gladier.tools.hello_world.HelloWorld')

        :param tool_ref: A tool ref can be a dotted import string or an actual GladierBaseTool
                         class.
        :return: gladier.GladierBaseTool
        """
        log.debug(f'Looking for Gladier tool: {tool_ref} ({type(tool_ref)})')
        if isinstance(tool_ref, str):
            default_cls = gladier.utils.dynamic_imports.import_string(tool_ref)
            _, alias = gladier.utils.dynamic_imports.parse_alias(tool_ref)
            default_inst = default_cls(alias, alias_class)
            if issubclass(type(default_inst), gladier.base.GladierBaseTool):
                return default_inst
            raise gladier.exc.ConfigException(f'{default_inst} is not of type '
                                              f'{gladier.base.GladierBaseTool}')
        elif isinstance(tool_ref, gladier.base.GladierBaseTool):
            return tool_ref
        else:
            cls_inst = tool_ref()
            if isinstance(cls_inst, gladier.base.GladierBaseTool):
                return cls_inst
            raise gladier.exc.ConfigException(
                f'"{tool_ref}" must be a {gladier.base.GladierBaseTool} or a dotted import '
                'string ')

    @property
    def version(self):
        return gladier.version.__version__

    @property
    def tools(self):
        """
        Load the current list of tools configured on this class

        :return: a list of subclassed instances of gladier.GladierBaseTool
        """
        if getattr(self, '_tools', None):
            return self._tools

        if not getattr(self, 'gladier_tools', None) or not isinstance(self.gladier_tools, Iterable):
            raise gladier.exc.ConfigException(
                '"gladier_tools" must be a defined list of Gladier Tools. '
                'Ex: ["gladier.tools.hello_world.HelloWorld"]')
        self._tools = [self.get_gladier_defaults_cls(gt, self.alias_class)
                       for gt in self.gladier_tools]
        return self._tools

    @property
    def scopes(self):
        """
        The current list of scopes required by this class. This changes if there
        is a flow configured in the local Gladier config file, otherwise it will
        only consist of basic scopes for running the funcx client/flows client/etc

        :return: list of globus scopes required by this client
        """
        return list(self.login_manager.get_authorizers().keys())

    @property
    def missing_authorizers(self):
        """
        :return:  a list of Globus scopes for which there are no authorizers
        """
        return self.login_manager.missing_authorizers

    def login(self):
        """
        Call ``login()`` on the configured login manager. Attepmts to prepare the user to run flows,
        but may require being called twice if a flow is not yet deployed. Automatically called
        internally by ``run_flow()`` if required.
        """
        if self.login_manager.missing_authorizers:
            self.login_manager.login(self.login_manager.missing_authorizers)

    def logout(self):
        """Call ``logout()`` on the login manager, to revoke saved tokens and deactivate the
        current flow. The flow_id and function ids/checksums are unaffected, and can be re-used
        after another invocation of ``login()``.
        """
        return self.login_manager.logout()

    def is_logged_in(self):
        """
        Check if the client is logged in.

        :return: True, if there are no self.missing_authorizers. False otherwise.
        """
        return self.login_manager.is_logged_in()

    def get_flow_definition(self):
        """
        Get the flow definition attached to this class. If the flow definition is an import string,
        it will automatically load the import string and return the full flow.

        :return: A dict of the Automate Flow definition
        """
        if not getattr(self, 'flow_definition', None):
            raise gladier.exc.ConfigException(f'"flow_definition" was not set on '
                                              f'{self.__class__.__name__}')

        if isinstance(self.flow_definition, dict):
            return self.flow_definition
        elif isinstance(self.flow_definition, str):
            return self.get_gladier_defaults_cls(self.flow_definition).flow_definition
        raise gladier.exc.ConfigException('"flow_definition" must be a dict or an import string '
                                          'to a sub-class of type '
                                          '"gladier.GladierBaseTool"')

    def sync_flow(self):
        self.flows_manager.flow_definition = self.get_flow_definition()
        self.flows_manager.sync_flow()

    def run_flow(self, flow_input=None, use_defaults=True, **flow_kwargs):
        combine_flow_input = self.get_input() if use_defaults else dict()
        if flow_input is not None:
            if not flow_input.get('input') or len(flow_input.keys()) != 1:
                raise gladier.exc.ConfigException(
                    f'Malformed input to flow, all input must be nested under "input", got '
                    f'{flow_input.keys()}')
            combine_flow_input['input'].update(flow_input['input'])
        for tool in self.tools:
            self.check_input(tool, combine_flow_input)

        self.sync_flow()
        return self.flows_manager.run_flow(flow_input=combine_flow_input, **flow_kwargs)

    def get_funcx_function_ids(self):
        """Get all funcx function ids for this run, registering them if there are no ids
        stored in the local Gladier config file OR the stored function id checksums do
        not match the actual functions provided on each of the Gladier tools. If register
        is False, no changes to the config will be made and exceptions will be raised instead.

        :raises: gladier.exc.RegistrationException
        :raises: gladier.exc.FunctionObsolete
        :returns: a dict of function ids where keys are names and values are funcX function ids.
        """
        funcx_ids = dict()
        for tool in self.tools:
            log.debug(f'Checking functions for {tool}')
            funcx_funcs = getattr(tool, 'funcx_functions', [])
            if not funcx_funcs:
                log.warning(f'Tool {tool} did not define any funcX functions!')
            if not funcx_funcs and not isinstance(funcx_funcs, Iterable):
                raise gladier.exc.DeveloperException(
                    f'Attribute "funcx_functions" on {tool} needs to be an iterable! Found '
                    f'{type(funcx_funcs)}')

            for func in funcx_funcs:
                name, val = self.funcx_manager.validate_function(tool, func)
                funcx_ids[name] = val
        return funcx_ids

    def get_flow_id(self) -> str:
        """
        Get the flow id from the flows manager class.
        """
        return self.flows_manager.get_flow_id()

    def get_input(self) -> dict:
        """
        Get funcx function ids, funcx endpoints, and each tool's default input. Default
        input may not be enough to run the flow. For example if a tool does processing on a
        local filesystem, the file will always need to be provided by the user when calling
        run_flow().

        Defaults rely on GladierBaseTool.flow_input defined separately for each tool.

        :return: input for a flow wrapped in an 'input' dict. For example:
                 {'input': {'foo': 'bar'}}
        """
        flow_input = self.get_funcx_function_ids()
        for tool in self.tools:
            # conflicts = set(flow_input.keys()).intersection(set(tool.flow_input))
            # if conflicts:
            #     for prev_tools in tools:
            #         for r in prev_tools.flow_input:
            #             if set(flow_input.keys()).intersection(set(tool.flow_input)):
            #                 raise gladier.exc.ConfigException(
            #                   f'Conflict: Tools {tool} and {prev_tool} 'both define {r}')
            flow_input.update(tool.get_flow_input())
            # Iterate over both private and public input variables, and include any relevant ones
            # Note: Precedence starts and ends with: Public --> Private --> Default on Tool
            input_keys = set(tool.get_flow_input()) | set(tool.get_required_input())
            log.debug(f'{tool}: Looking for overrides for the following input keys: {input_keys}')
            override_values = {k: self.storage.get_value(k) for k in input_keys
                               if self.storage.get_value(k) is not None}
            if override_values:
                log.info(f'Updates from {self.storage.filename}: {list(override_values.keys())}')
                flow_input.update(override_values)
        return {'input': flow_input}

    def check_input(self, tool: GladierBaseTool, flow_input: dict):
        """
        Do basic checking on included input against requirements set by a tool. Raises an
        exception if the check does not 'pass'

        :param tool: The gladier.GladierBaseTool tool set in self.tools
        :param flow_input: Flow input intended to be passed to run_flow()
        :raises: gladier.exc.ConfigException
        """
        for req_input in tool.get_required_input():
            if req_input not in flow_input['input']:
                raise gladier.exc.ConfigException(
                    f'{tool} requires flow input value: "{req_input}"')

    def get_status(self, action_id: str):
        """
        Get the current status of the automate flow. Attempts to do additional work on funcx
        functions to deserialize any exception output.

        :param action_id: The globus action UUID used for this flow. The Automate flow id is
                          always the flow_id configured for this tool.
        :raises: Globus Automate exceptions from self.flows_client.flow_action_status
        :returns: a Globus Automate status object (with varying state structures)
        """
        return self.flows_manager.get_status(action_id)

    def progress(self, action_id, callback=None):
        """
        Continuously call self.get_status() until the flow completes. Each status response is
        used as a parameter to the provided callback, by default will use the builtin callback
        to print the current state to stdout.

        :param action_id: The action id for a running flow. The flow is automatically pulled
                          based on the current tool's flow_definition.
        :param callback: The function to call with the result from self.get_status. Must take
                         a single parameter: mycallback(self.get_status())
        """
        return self.flows_manager.progress(action_id, callback=callback)

    def get_details(self, action_id, state_name):
        """
        Attempt to extrapolate details from get_status() for a given state_name define in the flow
        definition. Note: This is usually only possible when a flow completes.

        :param action_id: The action_id for this flow. Flow id is automatically determined based
                          on the current tool being run.
        :param state_name: The state in the automate definition to fetch
        :returns: sub-dict of get_status() describing the :state_name:.
        """
        return gladier.utils.automate.get_details(self.get_status(action_id), state_name)



from importlib import import_module


def parse_alias(dotted_string_path):
    if ':' in dotted_string_path:
        return dotted_string_path.rsplit(':', 1)
    return dotted_string_path, None


def import_string(dotted_string_path):
    dotted_string_path, _ = parse_alias(dotted_string_path)
    module_paths = dotted_string_path.rsplit('.', 1)
    if len(module_paths) != 2:
        raise ImportError(f'Path "{dotted_string_path} must be a dotted path')
    module_path, class_name = module_paths
    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as ae:
        raise ImportError(f'Failed to import {class_name} from {module}.') from ae



import logging
from gladier.exc import FlowModifierException
from gladier.utils.name_generation import get_funcx_flow_state_name

log = logging.getLogger(__name__)
funcx_modifiers = {'endpoint', 'payload', 'tasks'}
# All top level states can be modified.
# https://globus-automate-client.readthedocs.io/en/latest/authoring_flows.html#action-state-type
state_modifiers = {
    'Type',
    'ActionUrl',
    'WaitTime',
    'ExceptionOnActionFailure',
    'RunAs',
    'InputPath',
    'Parameters',
    'ResultPath',
    'Catch',
    'ActionScope',
    'Next',
    'End',
}


class FlowModifiers:
    supported_modifiers = state_modifiers.union(funcx_modifiers)
    funcx_modifiers = funcx_modifiers
    state_modifiers = state_modifiers

    def __init__(self, tools, modifiers, cls=None):
        self.cls = cls
        self.tools = tools
        self.functions = [func for tool in tools for func in tool.funcx_functions]
        self.function_names = [f.__name__ for f in self.functions]
        self.state_names = [get_funcx_flow_state_name(f) for f in self.functions]
        self.modifiers = modifiers
        self.check_modifiers()

    def get_function(self, function_identifier):
        if function_identifier in self.function_names:
            return self.functions[self.function_names.index(function_identifier)]
        if function_identifier in self.functions:
            return function_identifier

    def get_flow_state_name(self, function_identifier):
        func = self.get_function(function_identifier)
        return get_funcx_flow_state_name(func)

    def get_state_result_path(self, state_name):
        return f'$.{state_name}.details.results'

    def check_modifiers(self):
        log.debug(f'Checking modifiers: {self.modifiers}')
        if not isinstance(self.modifiers, dict):
            raise FlowModifierException(f'{self.cls}: Flow Modifiers must be a dict')

        # Check if modifiers were set correctly
        for name, mods in self.modifiers.items():
            if not self.get_function(name):
                raise FlowModifierException(f'Class {self.cls} included modifier which does not '
                                            f'exist: {name}. Allowed modifiers include '
                                            f'{", ".join(self.function_names)}')

            for mod_name, mod_value in mods.items():
                if mod_name not in self.supported_modifiers:
                    raise FlowModifierException(f'Class {self.cls}: Unsupported modifier '
                                                f'"{mod_name}". The only supported modifiers are: '
                                                f'{self.supported_modifiers}')

    def apply_modifiers(self, flow):
        for name, mods in self.modifiers.items():
            state_name = self.get_flow_state_name(name)
            flow['States'][state_name] = self.apply_modifier(flow['States'][state_name], mods)
        return flow

    def apply_modifier(self, flow_state, state_modifiers):

        for modifier_name, value in state_modifiers.items():
            log.debug(f'Applying modifier "{modifier_name}", value "{value}"')
            # If this is for a funcx task
            if modifier_name in self.funcx_modifiers:
                if modifier_name == 'tasks':
                    flow_state['Parameters'] = self.generic_set_modifier(
                        flow_state['Parameters'], 'tasks', value)
                else:
                    flow_state['Parameters']['tasks'] = [
                        self.generic_set_modifier(fx_task, modifier_name, value)
                        for fx_task in flow_state['Parameters']['tasks']
                    ]
            elif modifier_name in self.state_modifiers:
                self.generic_set_modifier(flow_state, modifier_name, value)
        return flow_state

    def generic_set_modifier(self, item, mod_name, mod_value):
        if not isinstance(mod_value, str):
            if mod_value in self.functions:
                sn = get_funcx_flow_state_name(mod_value)
                mod_value = self.get_state_result_path(sn)
        elif isinstance(mod_value, str) and not mod_value.startswith('$.'):
            if mod_value in self.function_names:
                sn = self.state_names[self.function_names.index(mod_value)]
                mod_value = self.get_state_result_path(sn)
            elif mod_value in self.state_names:
                mod_value = self.get_state_result_path(mod_value)
            elif mod_name not in state_modifiers:
                mod_value = f'$.input.{mod_value}'

        # Remove duplicate keys
        for duplicate_mod_key in (mod_name, f'{mod_name}.$'):
            if duplicate_mod_key in item.keys():
                item.pop(duplicate_mod_key)

        # Note: Top level State types don't end with '.$', all others must end with
        # '.$' to indicate the value should be replaced. '.=' is not supported or possible yet
        if isinstance(mod_value, str) and mod_value.startswith('$.'):
            mod_name = f'{mod_name}.$'
        item[mod_name] = mod_value
        log.debug(f'Set modifier {mod_name} to {mod_value}')
        return item



import logging
import hashlib

from gladier.base import GladierBaseTool
from gladier.managers.service_manager import ServiceManager
from funcx import FuncXClient
from funcx.serialize import FuncXSerializer

import gladier.utils.funcx_login_manager

log = logging.getLogger(__name__)


class FuncXManager(ServiceManager):

    def __init__(self, auto_registration: bool = True, **kwargs):
        super().__init__(**kwargs)
        self.auto_registration = auto_registration

    def get_scopes(self):
        return gladier.utils.funcx_login_manager.FuncXLoginManager.SCOPES

    @property
    def funcx_client(self):
        """
        :return: an authorized funcx client
        """
        if getattr(self, '__funcx_client', None) is not None:
            return self.__funcx_client

        funcx_login_manager = gladier.utils.funcx_login_manager.FuncXLoginManager(
            authorizers=self.login_manager.get_manager_authorizers()
        )

        self.__funcx_client = FuncXClient(login_manager=funcx_login_manager)
        return self.__funcx_client

    @staticmethod
    def get_funcx_function_name(funcx_function):
        """
        Generate a function name given a funcx function. These function namse are used to refer
        to funcx functions within the config. There is no guarantee of uniqueness for function
        names.

        :return: human readable string identifier for a function (intended for a gladier.cfg file)
        """
        return f'{funcx_function.__name__}_funcx_id'

    @staticmethod
    def get_funcx_function_checksum(funcx_function):
        """
        Get the SHA256 checksum of a funcx function
        :return: sha256 hex string of a given funcx function
        """
        fxs = FuncXSerializer()
        serialized_func = fxs.serialize(funcx_function).encode()
        return hashlib.sha256(serialized_func).hexdigest()

    def validate_function(self, tool: GladierBaseTool, function):
        fid_name = gladier.utils.name_generation.get_funcx_function_name(function)
        fid = self.storage.get_value(fid_name)
        checksum = self.get_funcx_function_checksum(function)
        checksum_name = gladier.utils.name_generation.get_funcx_function_checksum_name(function)
        try:
            if not fid_name:
                raise gladier.exc.RegistrationException(
                    f'Tool {tool.__class__.__name__} missing funcx registration for {fid_name}')
            if not self.storage.get_value(checksum_name):
                raise gladier.exc.RegistrationException(
                    f'Tool {tool.__class__.__name__} with function {fid_name} '
                    f'has a function id but no checksum!')
            if not self.storage.get_value(checksum_name) == checksum:
                raise gladier.exc.FunctionObsolete(
                    f'Tool {tool.__class__.__name__} with function {fid_name} '
                    f'has changed and needs to be re-registered.')
        except (gladier.exc.RegistrationException, gladier.exc.FunctionObsolete):
            if self.auto_registration is True:
                log.info(f'{tool.__class__.__name__}: function {function.__name__} is out of date')
                fid = self.register_function(tool, function)
                fx_name = gladier.utils.name_generation.get_funcx_function_name(function)
                self.storage.set_value(fx_name, fid)
                self.storage.set_value(checksum_name, checksum)
            else:
                raise
        finally:
            return fid_name, fid

    def register_function(self, tool: GladierBaseTool, function):
        """Register the functions with funcx. Ids are saved in the local gladier.cfg"""
        log.info(f'{tool.__class__.__name__}: registering function {function.__name__}')
        return self.funcx_client.register_function(function)



import functools
from gladier.base import GladierBaseTool
from gladier.client import GladierBaseClient
from gladier.exc import FlowGenException
from gladier.utils.flow_generation import generate_tool_flow, combine_tool_flows


def generate_flow_definition(_cls=None, *, modifiers=None):
    """Class decorators for automatically generating flows on either
    GladierBaseTools or GladierClients. For GladierBaseTools, this generates
    a simple flow containing all attached funcx_functions, applying any modifiers
    for custom values the GladierBaseTool needs to run. For GladierClients, this
    instead stitches together all flows defined on GladierBaseTools, in the order
    they are defined for each BaseTool. ``modifiers`` are only allowed on
    GladierBaseTools.

    Example:
    @generate_flow_definition(modifiers={
        my_funcx_function: {
            'payload': '$.MyFuncxFunction.details.result',
            'endpoint': 'funcx_endpoint_non_compute'
        }
    })

    Any modifier values without a preceding '$.' will be replaced with
    a path to general input. The above would result in the following
    for ``funcx_endpoint_non_compute`` above:
        '$.input.funcx_endpoint_non_compute'

    :raises FlowGenException: For a variety of invalid inputs"""
    modifiers = modifiers or dict()

    def decorator_wrapper(cls):
        @functools.wraps(cls)
        def wrapper(*args, **kwargs):
            if issubclass(cls, GladierBaseTool):
                c = cls(*args, **kwargs)
                c.flow_definition = generate_tool_flow(c, modifiers)
                return c
            elif issubclass(cls, GladierBaseClient):
                c = cls(*args, **kwargs)
                c.flow_definition = combine_tool_flows(c, modifiers)
                return c
            else:
                raise FlowGenException(f'Invalid class {cls}, flow generation '
                                       f'only supported for '
                                       f'{[GladierBaseTool, GladierBaseClient]}')

        return wrapper

    if _cls is None:
        return decorator_wrapper
    else:
        return decorator_wrapper(_cls)



import logging
from typing import Mapping, Any, List
import copy

from gladier.base import GladierBaseTool
from gladier.exc import FlowGenException, StateNameConflict
from gladier.utils.flow_traversal import get_end_states

log = logging.getLogger(__name__)


class ToolChain:

    def __init__(self, flow_comment=None):
        self._flow_definition = {
            'States': dict(),
            'Comment': flow_comment,
            'StartAt': None,
        }
        self.transition_states = list()

    @property
    def flow_definition(self):
        flow_def = copy.deepcopy(self._flow_definition)
        if not self._flow_definition.get('Comment'):
            state_names = ", ".join(flow_def["States"].keys())
            flow_def['Comment'] = f'Flow with states: {state_names}'
            return flow_def
        return flow_def

    def chain(self, tools: List[GladierBaseTool]):
        """
        Given a list of tools, chain each one to the existing flow definition
        stored on this class. If there is no current flow definition, the first
        flow in the first tool will be used as the starting point.
        :raises FlowGenException: if there was a problem chaining together flows
        :returns self: a reference to this class.
        """
        self.check_tools(tools)
        for tool in tools:
            log.debug(f'Chaining tool {tool.__class__.__name__} to existing flow '
                      f'({len(self._flow_definition["States"])} states)')

            flow_definition = tool.get_flow_definition()
            self._chain_flow(flow_definition, tool)

        return self

    def chain_state(self, name: str, definition: Mapping[str, Any]):
        """Chain a single flow state to the existing flow definition. If there
        is no existing flow definition, a new flow definition will be created."""
        log.debug(f'Chaining state {name} with definition {definition.keys()}')
        temp_flow = {
            'StartAt': name,
            'States': {name: definition},
        }
        temp_flow['States'][name]['End'] = True
        self._chain_flow(temp_flow)
        return self

    def _chain_flow(self, new_flow: Mapping[str, dict], tool: GladierBaseTool = None):
        # Base case, if this is the first 'chain' and no states exist yet.
        if not self._flow_definition['States']:
            self._flow_definition['States'] = copy.deepcopy(new_flow['States'])
            self._flow_definition['StartAt'] = new_flow['StartAt']
        else:
            self._flow_definition['States'].update(copy.deepcopy(new_flow['States']))
            for t_state in self.transition_states:
                self.add_transition(t_state, new_flow['StartAt'])

        self._set_new_transition_states(new_flow, tool)

    def _set_new_transition_states(self, new_flow: Mapping[str, dict],
                                   tool: GladierBaseTool = None):
        # Use the tool-defined states if they are defined, otherwise there is only one
        # End state on the flow and therefore can be assumed.
        t_states = list(get_end_states(new_flow))
        if tool:
            tool_t_states = tool.get_flow_transition_states()
        else:
            tool_t_states = []
        entity = tool or f'Flow with states: {tuple(new_flow["States"].keys())}'
        if len(t_states) > 1 and not tool_t_states:
            raise FlowGenException(f'{entity} has multiple branching end states and must '
                                   'define which states a flow may continue by setting '
                                   '"flow_transition_states" containing one or more of '
                                   f'{", ".join(t_states)} ')
        elif not t_states:
            raise FlowGenException(f'{tool}: Could not find any end states in flow.')

        self.transition_states = tool_t_states if tool_t_states else t_states

    def add_transition(self, cur_flow_term: str, new_chain_start: str):
        if self._flow_definition['States'][cur_flow_term].get('End'):
            self._flow_definition['States'][cur_flow_term].pop('End')
        log.debug(f'Chaining {cur_flow_term} --> {new_chain_start}')
        self._flow_definition['States'][cur_flow_term]['Next'] = new_chain_start

    def check_tools(self, tools: List[GladierBaseTool]):

        states = set()
        for tool in tools:
            flow_def = tool.get_flow_definition()
            if flow_def is None:
                raise FlowGenException(f'Tool {tool} did not set .flow_definition attribute or set '
                                       f'@generate_flow_definition (funcx functions only). Please '
                                       f'set a flow definition for {tool.__class__.__name__}.')

            new_states = set(flow_def.get('States'))
            conflicts = states & new_states
            if conflicts:
                raise StateNameConflict(f'States in tool {tool} '
                                        f'has been defined more than once: {conflicts}')
            states = states | new_states



from abc import ABC, abstractmethod
from packaging import version
import logging
import gladier.version

log = logging.getLogger(__name__)


class ConfigMigration(ABC):

    def __init__(self, config):
        self.config = config

        if 'general' not in self.config.sections():
            log.debug('No version in config, adding...')
            self.config.add_section('general')
        cfg_version = self.config['general'].get('version')

        self.config_version = version.parse(cfg_version) if cfg_version else None
        self.version = version.parse(gladier.version.__version__)

    @abstractmethod
    def is_applicable(self):
        pass

    @abstractmethod
    def migrate(self):
        pass


class AddVersionToConfig(ConfigMigration):
    """
    Add a version if one does not exist in the config.
    """
    def is_applicable(self):
        return self.config_version is None

    def migrate(self):
        # We can't be certain which version of Gladier the user was using previously.
        # It's possible they simply upgraded from 3 to 4 and are using incompatible
        # funcx functions. Run the migration here just in case.
        migrate_delete_all_funcx_functions(self.config)

        # Set the version
        self.config['general']['version'] = str(self.version)
        log.info(f'Setting Version {self.version}')


class UpdateConfigVersion(ConfigMigration):
    """Sets the config version to the current version of Gladier.
    Should be run LAST, after all other migrations have taken place"""
    def is_applicable(self):
        return self.config_version != self.version

    def migrate(self):
        self.config['general']['version'] = str(self.version)


MIGRATIONS = [
    AddVersionToConfig,
    UpdateConfigVersion,
]


def needs_migration(config):
    return any(m(config).is_applicable() for m in MIGRATIONS)


def migrate_gladier(config):
    for migration in MIGRATIONS:
        m_instance = migration(config)
        if m_instance.is_applicable():
            log.info(f'Applying migration: {m_instance.__class__.__name__}')
            m_instance.migrate()
    return config


def panic_print(message):
    """Print a message to console for the user to see. The message must be URGENT."""
    print(message)


def migrate_delete_all_funcx_functions(config):
    """FuncX changed servers from v0.0.5 to v0.2.4/v0.3.0. Simply delete all functions
    to upgrade to the latest version. The new version of funcx will re-register them."""
    for section in config.sections():
        for option in config[section]:
            if option.endswith('_funcx_id') or option.endswith('_funcx_id_checksum'):
                log.debug(f'Deleting {option}')
                del config[section][option]



import logging
import configparser
from gladier.storage.migrations import needs_migration, migrate_gladier

log = logging.getLogger(__name__)


class GladierConfig(configparser.ConfigParser):

    def __init__(self, filename, section='default'):
        super().__init__()
        self.section = section
        self.filename = filename
        self.load()

    def load(self):
        self.read(self.filename)
        if self.section not in self.sections():
            log.debug(f'Section {self.section} missing, adding to config.')
            self[self.section] = {}
            self.save()

    def save(self):
        with open(self.filename, 'w') as configfile:
            self.write(configfile)
            log.debug(f'Saved local gladier config to {configfile}')

    def update(self):
        if needs_migration(self):
            self = migrate_gladier(self)
            self.save()

    def get_value(self, name: str) -> str:
        try:
            self.load()
            return self.get(self.section, name)
        except configparser.NoOptionError:
            return None

    def set_value(self, name: str, value: str):
        self.load()
        self.set(self.section, name, value)
        self.save()

    def del_value(self, name: str) -> None:
        self.load()
        self.remove_option(self.section, name)
        self.save()



import logging
import typing
from gladier.base import GladierBaseTool
from gladier.client import GladierBaseClient
from gladier.exc import FlowGenException
from gladier.utils.flow_modifiers import FlowModifiers
from gladier.utils.name_generation import (
    get_funcx_flow_state_name,
    get_funcx_function_name
)
from gladier.utils.tool_chain import ToolChain


log = logging.getLogger(__name__)


def combine_tool_flows(client: GladierBaseClient, modifiers):
    """
    Combine flow definitions on each of a Gladier Client's **tools** and return
    a single flow definition that runs each state in order from first to last.

    Modifiers can be applied to any of the states within the flow.
    """
    flow_moder = FlowModifiers(client.tools, modifiers, cls=client)
    chain = ToolChain(flow_comment=client.__doc__).chain(client.tools)
    return flow_moder.apply_modifiers(chain.flow_definition)


def _get_duplicate_functions(funcx_functions: typing.List[callable]):
    tracked_set = set()
    func_names = [get_funcx_flow_state_name(f) for f in funcx_functions]
    return [f for f in func_names if f in tracked_set or tracked_set.add(f)]


def generate_tool_flow(tool: GladierBaseTool, modifiers):
    """Generate a flow definition for a Gladier Tool based on the defined ``funcx_functions``.
    Accepts modifiers for funcx functions"""
    duplicate_functions = _get_duplicate_functions(tool.funcx_functions)
    if duplicate_functions:
        raise FlowGenException(f'Tool {tool} contains duplicate function names: '
                               f'{duplicate_functions}')

    flow_moder = FlowModifiers([tool], modifiers, cls=tool)

    tools = ToolChain()
    for fx_func in tool.funcx_functions:
        tools.chain_state(*generate_funcx_flow_state(fx_func))

    flow = tools.flow_definition
    if not flow['States']:
        raise FlowGenException(f'Tool {tool} has no flow states. Add a list of python functions '
                               f'as "{tool}.funcx_functions = [myfunction]" or set a custom flow '
                               f'definition instead using `{tool}.flow_definition = mydef`')
    return flow_moder.apply_modifiers(flow)


def generate_funcx_flow_state(funcx_function):
    state_name = get_funcx_flow_state_name(funcx_function)

    return state_name, {
        'Comment': funcx_function.__doc__,
        'Type': 'Action',
        'ActionUrl': 'https://automate.funcx.org',
        'ActionScope': 'https://auth.globus.org/scopes/'
                       'b3db7e59-a6f1-4947-95c2-59d6b7a70f8c/action_all',
        'ExceptionOnActionFailure': False,
        'Parameters': {
            'tasks': [{
                'endpoint.$': '$.input.funcx_endpoint_compute',
                'function.$': f'$.input.{get_funcx_function_name(funcx_function)}',
                'payload.$': '$.input',
            }]
        },
        'ResultPath': f'$.{state_name}',
        'WaitTime': 300,
    }



import logging
from gladier.client import GladierBaseClient
from gladier.base import GladierBaseTool
from gladier.decorators import generate_flow_definition
from gladier.managers import FlowsManager, CallbackLoginManager

__all__ = [
    GladierBaseClient, GladierBaseTool, generate_flow_definition, FlowsManager,
    CallbackLoginManager
]

# https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library  # noqa
logging.getLogger("gladier").addHandler(logging.NullHandler())



import logging
from typing import Callable, List, Set, Iterable, Union, Any, Mapping
import abc

import fair_research_login
from globus_sdk import AccessTokenAuthorizer, RefreshTokenAuthorizer
from gladier.exc import AuthException

log = logging.getLogger(__name__)


class BaseLoginManager(abc.ABC):

    def __init__(self, *args, **kwargs):
        self.required_scopes = set()
        self.scope_changes = set()

    @property
    def missing_authorizers(self) -> Set[str]:
        return self.get_missing_authorizers(self.get_authorizers())

    def get_missing_authorizers(
        self,
        authorizers: Mapping[str, Union[AccessTokenAuthorizer, RefreshTokenAuthorizer]]
            ) -> Set[str]:
        # Disregard any scopes not in the 'required' list. This allows implementers to return
        # unrelated scopes.
        absent = self.required_scopes.difference(authorizers)
        log.info(f'Scopes Absent: {absent or None}, Need Update: {self.scope_changes or None}')
        return absent | self.scope_changes

    def get_manager_authorizers(self):
        authorizers = self.get_authorizers()
        missing = self.get_missing_authorizers(authorizers)

        if missing:
            log.info('Attempting login to fetch missing authorizers.')
            self.login(missing)
            authorizers = self.get_authorizers()
            missing = self.get_missing_authorizers(authorizers)

        if missing:
            raise AuthException('Manager Failed to produce '
                                f'authorizers for missing scopes: {missing}')
        return authorizers

    @abc.abstractmethod
    def get_authorizers(self) -> Mapping[str, Union[AccessTokenAuthorizer, RefreshTokenAuthorizer]]:
        """
        This method MUST be implemented by the implementing class.

        Load any existing authorizers and return them to the client. A login *should not* be
        attempted. The Login Manager will determine if the scopes provided are sufficient, and
        automatically call login() they are not.
        """
        raise NotImplementedError(f'{self.__class__.__name__} must implement .get_authorizers()')

    @abc.abstractmethod
    def login(self, scopes: List[str]):
        """
        This method MUST be implemented by the implementing class.

        If this method is called, get_authorizers() has insufficient scopes and requires more
        scopes to continue operations. ``scopes`` MAY reference scopes already saved, which can
        happen if dependent scopes have been added to the scope since last login. For this reason,
        previously saved scopes in this list should be disregarded.
        """
        raise NotImplementedError(f'{self.__class__.__name__} must implement .ensure_logged_in()')

    def is_logged_in(self) -> bool:
        return not bool(self.missing_authorizers)

    def add_requirements(self, scopes: Iterable[str]):
        log.debug(f'Added required scopes {scopes}')
        self.required_scopes = self.required_scopes | set(scopes)

    def add_scope_change(self, scopes: Iterable[str]):
        """
        A scope change happens when the underlying scope gets updated with new dependent scopes,
        and the current token we now have is invalid and can no longer be used. The solution is
        a new login to get a new token.
        """
        log.debug(f'Tracking scope change: {scopes}')
        self.scope_changes = self.scope_changes | set(scopes)


class AutoLoginManager(BaseLoginManager):
    """
    Default Login Manager class in Gladier. Automatically initiates a login if any scope
    is missing or needs to be updated.
    """

    refresh_tokens = True

    def __init__(self, client_id: str, storage: Any, app_name: str, auto_login=True):
        super().__init__()
        self.storage = storage
        self.client_id = client_id
        self.app_name = app_name
        self.auto_login = auto_login

    @property
    def native_client(self):
        if getattr(self, 'client_id', None) is None:
            raise AuthException(
                'Gladier client must be instantiated with a '
                '"client_id" to use "login()!'
            )
        return fair_research_login.NativeClient(client_id=self.client_id,
                                                app_name=self.app_name,
                                                token_storage=self.storage)

    def get_authorizers(self) -> List[Union[AccessTokenAuthorizer, RefreshTokenAuthorizer]]:
        try:
            return self.native_client.get_authorizers_by_scope()
        except fair_research_login.LoadError:
            return dict()

    def login(self, scopes):
        if self.auto_login is False:
            raise AuthException(f'Automatic login is disabled. Missing scopes: {scopes}')
        self.native_client.login(requested_scopes=scopes,
                                 refresh_toknes=self.refresh_tokens,
                                 force=True)

    def logout(self):
        self.native_client.logout()


class CallbackLoginManager(BaseLoginManager):
    """
    The Callback Login Manager allows for finer grained control of auth within Gladier. Logins
    are lazy up until Gladier attempts to make a call to the FuncX service, Flows service, or
    tries to run a flow. Scopes for a deployed flow may be modified at any time, requiring
    a re-login with that flow scope.

    :param authorizers: A dict of autorizers by scope. If these satisfy the login requirements
                        then the callback will not be triggered.
    :param callback: A Callable which will be invoked if additional scopes are needed.

    """

    def __init__(
        self,
        authorizers: Mapping[str, Union[AccessTokenAuthorizer, RefreshTokenAuthorizer]],
        callback: Callable = None
    ):
        super().__init__()
        self.authorizers = authorizers
        self.callback = callback

    def get_authorizers(self) -> List[Union[AccessTokenAuthorizer, RefreshTokenAuthorizer]]:
        return self.authorizers

    def login(self, scopes):
        if not self.callback:
            raise AuthException('New login required for scopes and no callback set on '
                                f'login manager: {scopes}')
        new_authorizers = self.callback(scopes)
        self.authorizers.update(new_authorizers)
        log.info('New authorizers have been cached for re-use.')



import logging
import warnings
import hashlib
import json
from typing import Callable

import globus_sdk
import globus_automate_client

import gladier
from gladier.managers.service_manager import ServiceManager
import gladier.storage.config
import gladier.utils.dynamic_imports
import gladier.utils.automate
import gladier.utils.name_generation
import gladier.storage.migrations
import gladier.utils.tool_alias
import gladier.utils.funcx_login_manager
import gladier.exc
import gladier.version

from globus_automate_client.flows_client import (
    MANAGE_FLOWS_SCOPE,
    VIEW_FLOWS_SCOPE,
    RUN_FLOWS_SCOPE,
    RUN_STATUS_SCOPE,
    RUN_MANAGE_SCOPE,
)

log = logging.getLogger(__name__)


def ensure_flow_registered(flows_manager_instance, exc: gladier.exc.RegistrationException):
    log.info(f'Deploying/Updating flow due to: {str(exc)}')
    flows_manager_instance.register_flow()


class FlowsManager(ServiceManager):
    """
    The flows manager tracks an externally defined flow_definition and ensures it stays
    up-to-date with a deployment in the flows service. It can be run without a flow_id
    in which case it will deploy its own flow if a stored flow_id does not exist.

    The flow checksum is evaluated inside ``run_flow()``, and any updates to the flow
    will take place before the flow is started. The registration behavior can be customized
    by passing in a function for ``on_change``. Default behavior will call ``register_flow()``,
    None will disable changing the flow before starting it. The ``on_change`` signature is below:

    .. code-block::

        def on_change_callback(flows_manager_instance: FlowsManager,
                               exc: gladier.exc.RegistrationException) -> None:

    :param flow_id: Explicit flow id to use. None will result in deploying a new flow
    :param flow_definition: Flow definiton that should be used. Usually set dynamically
                            at runtime when used with a Gladier Client
    :param flow_schema: The schema to be used alongside the flow definition
    :param flow_title: The title for the Globus Flow
    :param globus_group: A Globus Group UUID. Used to grant all flow and run permissions
    :param subscription_id: (deprecated) Subscription ID has no effect and will be removed in a
                            future version.
    :param on_change: callback on checksum mismatch or missing flow id. Default registers/deploys
                      flow, ``None`` takes no action and attempts to run "obselete" flows.
    :param redeploy_on_404: Deploy a new flow if attempting to run the current flow ID results in 404.
                            Behavior is disabled if an explicit flow_id is specified.

    When used with a Gladier Client, following items will be auto-configured and should not be
    set explicitly in the constructor:

     * flow_definition
     * flow_schema


    .. note::

        The FlowsManager class cannot be used to run flows outside of Gladier Clients due to
        internal class storage requirements. Doing so will result in an exception.
    """

    AVAILABLE_SCOPES = [
        MANAGE_FLOWS_SCOPE,
        VIEW_FLOWS_SCOPE,
        RUN_FLOWS_SCOPE,
        RUN_STATUS_SCOPE,
        RUN_MANAGE_SCOPE,
    ]

    def __init__(self,
                 flow_id: str = None,
                 flow_definition: dict = None,
                 flow_schema: dict = None,
                 flow_title: str = None,
                 globus_group: str = None,
                 subscription_id: str = None,
                 on_change: Callable = ensure_flow_registered,
                 redeploy_on_404: bool = True,
                 **kwargs):
        self.flow_id = flow_id
        self.flow_definition = flow_definition
        self.flow_schema = flow_schema
        self.flow_title = flow_title
        self.globus_group = globus_group
        self.on_change = on_change or (lambda self, exc: None)
        self.redeploy_on_404 = redeploy_on_404

        if subscription_id:
            warnings.warn(f'{self.flow_title}: subscription_id is no longer used and '
                          'should not be set on flows.', category=DeprecationWarning)

        if self.flow_id is not None:
            self.redeploy_on_404 = False
            log.info('Custom Flow ID set, redeploy_on_404 disabled.')

        super().__init__(**kwargs)

    def get_scopes(self):
        scopes = self.AVAILABLE_SCOPES.copy()
        flow_scope = self.flow_scope
        if flow_scope:
            scopes.append(flow_scope)
        return scopes

    @property
    def flow_scope(self):
        """Get the scope for a flow
        :returns: None if no flow id exists"""
        try:
            flow_id = self.get_flow_id()
        except AttributeError:
            # It's possible storage wasn't set yet, and get_flow_id() raises
            # an error here. Simply ignore it for the time being.
            flow_id = None
        if flow_id:
            # In the future, this should be gettable via
            # globus_sdk.SpecificFlowClient(flow_id).scopes.url_scope_string('flow_id'))
            scope_name = f'flow_{flow_id.replace("-", "_")}_user'
            return f'https://auth.globus.org/scopes/{flow_id}/{scope_name}'

    @property
    def flow_definition(self) -> dict:
        return self._flow_definition

    @flow_definition.setter
    def flow_definition(self, value: dict):
        self._flow_definition = value

    @property
    def flow_schema(self) -> dict:
        return self._flow_schema or {
            'input_schema': {
                'additionalProperties': True,
                'properties': {},
                'type': 'object'
            }
        }

    @flow_schema.setter
    def flow_schema(self, value: dict):
        self._flow_schema = value

    @property
    def flows_client(self):
        """
        :return: an authorized Gloubs Automate Client. If a flow has been deployed,
        the client returned will be authorized to run the flow. Stale clients pre-flow-deployment
        will need to call this again in order to run flows.
        """
        if getattr(self, '_flows_client', None) is not None:
            return self._flows_client
        authorizers = self.login_manager.get_manager_authorizers()

        automate_authorizer = authorizers[
            globus_automate_client.flows_client.MANAGE_FLOWS_SCOPE
        ]
        flow_authorizer = authorizers.get(self.flow_scope)

        def get_flow_authorizer(*args, **kwargs):
            return flow_authorizer

        self._flows_client = globus_automate_client.FlowsClient.new_client(
            None, get_flow_authorizer, automate_authorizer,
        )
        return self._flows_client

    def refresh_flows_client(self) -> globus_automate_client.FlowsClient:
        self._flows_client = None
        return self.flows_client

    @staticmethod
    def get_flow_checksum(flow_definition):
        """
        Get the SHA256 checksum of the current flow definition.

        :return: sha256 hex string of flow definition
        """
        return hashlib.sha256(json.dumps(flow_definition, sort_keys=True).encode()).hexdigest()

    @staticmethod
    def get_globus_urn(uuid, id_type='group'):
        """Convenience method for appending the correct Globus URN prefix on a uuid."""
        URN_PREFIXES = {
            'group': 'urn:globus:groups:id:',
            'identity': 'urn:globus:auth:identity:'
        }
        if id_type not in URN_PREFIXES:
            raise gladier.exc.DevelopmentException('"id_type" must be one of '
                                                   f'{URN_PREFIXES.keys()}. Got: {id_type}')
        return f'{URN_PREFIXES[id_type]}{uuid}'

    def get_flow_permission(self, permission_type, identities=None):
        """
        This function is a generic shim that should work for most Gladier clients that
        want basic permissions that will work with a single Globus Group. This method can be
        overridden to change any of the automate defaults:

        permission_type for deploying flows:
            'flow_viewers', 'flow_starters', 'flow_administrators',

        permission_type for running flows:
            'run_managers', 'run_monitors'

        By default, always returns either None for using automate defaults, or setting every
        permission_type above to use the set client `globus_group`.
        """
        if identities is None and self.globus_group:
            identities = [self.get_globus_urn(self.globus_group)]
        permission_types = {
            'flow_viewers', 'flow_starters', 'flow_administrators', 'run_managers', 'run_monitors'
        }
        if permission_type not in permission_types:
            raise gladier.exc.DevelopmentException(f'permission_type must be one of '
                                                   f'{permission_types}')
        return identities

    def get_flow_id(self) -> str:
        """Return flow id. If an ID was set on this class in the constructor, that is
        used. Otherwise, a retrieve from storage is attempted with 'flow_id'.
        :returns: flow_id uuid for deployed flow, or None if it does not exist"""
        return self.flow_id or self.storage.get_value('flow_id')

    def flow_changed(self) -> bool:
        """
        Returns True if a flow id exists and the stored checksum matches the set
        flow_definition. False otherwise.
        :raises: Nothing
        """
        try:
            self.check_flow()
        except gladier.exc.RegistrationException:
            return True
        return False

    def check_flow(self):
        """
        Check if the flow has changed by validating the current flow_definition agaist
        the stored checksum. Raises an exception if the checksums do not match.

        :raises: gladier.exc.NoFlowRegistered if no flow has been registered
        :raises: gladier.exc.FlowObsolete if the stored flow checksums do not match
        """
        flow_id = self.get_flow_id()
        flow_checksum = self.storage.get_value('flow_checksum')
        if not flow_id:
            raise gladier.exc.NoFlowRegistered(
                'No flow_id set on flow manager and no id tracked in storage.')
        elif flow_checksum != self.get_flow_checksum(self.flow_definition):
            raise gladier.exc.FlowObsolete(
                f'"flow_definition" on {self} has changed and needs to be re-registered.')

    def sync_flow(self) -> str:
        """
        Get the current flow id for the current Gladier flow definition.
        If self.auto_register is True, it will automatically (re)register a flow if it
        has changed on disk, otherwise raising exceptions.

        :param flow_definition: A complete deployable Flow Definition. Will replace existing
        flow definition if it differs.

        :raises: gladier.exc.FlowObsolete
        :raises: gladier.exc.NoFlowRegistered
        """
        try:
            self.check_flow()
        except gladier.exc.RegistrationException as exc:
            self.on_change(self, exc)
        return self.get_flow_id()

    def register_flow(self) -> str:
        """
        Deploy the current flow_definition. If a flow_id exists, the flow is updated
        instead. If the flow does not exist (404) and redeploy_on_404 is set, the flow will
        be automatically re-deployed with a new flow id.



        Note: If a new flow is deployed, or an existing scope adds unique Action Providers,
        a new login will be needed before the flow can be run.
        :raises: globus_sdk.exc.GlobusAPIError on error deploying flow
        :return: an automate flow UUID
        """
        flow_id = self.get_flow_id()
        flow_permissions = {
            p_type: self.get_flow_permission(p_type)
            for p_type in ['flow_viewers', 'flow_starters', 'flow_administrators']
            if self.get_flow_permission(p_type)
        }
        log.debug(f'Flow permissions set to: {flow_permissions or "Flows defaults"}')
        flow_kwargs = flow_permissions
        # Input schema is a required field and must be part of all flows.
        flow_kwargs['input_schema'] = self.flow_schema
        if flow_id:
            try:
                log.info(f'Flow checksum failed, updating flow {flow_id}...')
                self.flows_client.update_flow(flow_id, self.flow_definition, **flow_kwargs)
                self.storage.set_value('flow_checksum',
                                       self.get_flow_checksum(self.flow_definition))
            except globus_sdk.exc.GlobusAPIError as gapie:
                if gapie.http_status == 404 and self.redeploy_on_404:
                    flow_id = None
                else:
                    raise
        if flow_id is None:
            log.info('No flow detected, deploying new flow...')
            flow = self.flows_client.deploy_flow(self.flow_definition, title=self.flow_title,
                                                 **flow_kwargs).data
            self.storage.set_value('flow_id', flow['id'])
            self.storage.set_value('flow_checksum', self.get_flow_checksum(self.flow_definition))
            self.login_manager.add_requirements([flow['globus_auth_scope']])
            self.refresh_flows_client()

        return flow_id

    def purge_flow(self):
        """
        Remove the stored flow_id and flow_checksum.
        """
        self.storage.del_value('flow_id')
        self.storage.del_value('flow_checksum')

    def run_flow(self, **kwargs):
        r"""
        Start a Globus Automate flow. By default, the flow definiton is checked and synced if it
        has changed locally or deployed if it does not exist.

        If a group is set, run permissions are updated and applied to the run (includes
        'run_managers', 'run_monitors').

        Any scope changes required post-deployment/update are propogated through the login_manager
        and may require an additional login. A new flow checksum/id may be tracked in storage if
        the flow changed or was newly deployed.

        :param flow_input: A dict of input to be passed to the automate flow. self.check_input()
                           is called on each tool to ensure basic needs are met for each.
                           Input MUST be wrapped inside an 'input' dict,
                           for example {'input': {'foo': 'bar'}}.

        :param use_defaults: Use the result of self.get_input() to populate base input for the
                             flow. All conflicting input provided by flow_input overrides
                             values set in use_defaults.
        :param \**flow_kwargs: Set several keyed arguments that include the label to be used
                               in the automate app. If no label is passed the standard automate
                               label is used. Also ensure label <= 64 chars long.
        :raise: gladier.exc.ConfigException by self.check_input()
        :raises: gladier.exc.FlowObsolete
        :raises: gladier.exc.NoFlowRegistered
        :raises: gladier.exc.RegistrationException
        :raises: gladier.exc.FunctionObsolete
        :raises: gladier.exc.AuthException
        :raises: Any globus_sdk.exc.BaseException
        """
        flow_id = self.get_flow_id()

        permissions = {
            p_type: self.get_flow_permission(p_type)
            for p_type in ['run_managers', 'run_monitors']
            if self.get_flow_permission(p_type)
        }
        log.debug(f'Flow run permissions set to: {permissions or "Flows defaults"}')
        kwargs.update(permissions)

        # Ensure the label is not longer than 64 chars
        if 'label' in kwargs:
            label = kwargs['label']
            kwargs['label'] = (label[:62] + '..') if len(label) > 64 else label

        try:
            flow = self.flows_client.run_flow(flow_id, self.flow_scope, **kwargs).data
        except globus_sdk.exc.GlobusAPIError as gapie:
            log.debug('Encountered error when running flow', exc_info=True)
            automate_error_message = json.loads(gapie.message)
            detail_message = automate_error_message['error']['detail']
            if 'unable to get tokens for scopes' in detail_message:
                self.login_manager.add_scope_change([self.flow_scope])
                self.refresh_flows_client()
                log.info('Initiating new login for dependent scope change')
                flow = self.flows_client.run_flow(flow_id, self.flow_scope, **kwargs).data
            elif gapie.http_status == 404 and self.redeploy_on_404:
                log.warning(f'Flow {flow_id} returned 404 and is either deleted or unavailable. '
                            f'Purging flow_id from config file...')
                # On a 404, Gladier can't do anything since it cannot access the old flow.
                # Therefore, purge the old flow and deploy a new one.
                self.purge_flow()
                # Ensure the new flow is deployed
                self.sync_flow()
                # Fetch the new flow id
                flow_id = self.get_flow_id()
                flow = self.flows_client.run_flow(flow_id, self.flow_scope, **kwargs).data
            else:
                raise
        log.info(f'Started flow {kwargs.get("label")} flow id '
                 f'"{flow_id}" with run "{flow["run_id"]}"')

        if flow['status'] == 'FAILED':
            raise gladier.exc.ConfigException(f'Flow Failed: {flow["details"]["description"]}')
        return flow

    def get_status(self, run_id):
        """
        Get the current status of the automate flow. Attempts to do additional work on funcx
        functions to deserialize any exception output.

        :param run_id: The globus action UUID used for this flow. The Automate flow id is
                          always the flow_id configured for this tool.
        :raises: Globus Automate exceptions from self.flows_client.flow_action_status
        :returns: a Globus Automate status object (with varying state structures)
        """
        status = self.flows_client.flow_action_status(
                self.get_flow_id(), self.flow_scope, run_id
            ).data
        try:
            return gladier.utils.automate.get_details(status)
        except (KeyError, AttributeError):
            return status

    @staticmethod
    def _default_progress_callback(response):
        if response['status'] == 'ACTIVE':
            print(f'[{response["status"]}]: {response["details"]["description"]}')

    def progress(self, run_id, callback=None):
        """
        Continuously call self.get_status() until the flow completes. Each status response is
        used as a parameter to the provided callback, by default will use the builtin callback
        to print the current state to stdout.

        :param run_id: The action id for a running flow. The flow is automatically pulled
                          based on the current tool's flow_definition.
        :param callback: The function to call with the result from self.get_status. Must take
                         a single parameter: mycallback(self.get_status())
        """
        callback = callback or self._default_progress_callback
        status = self.get_status(run_id)
        while status['status'] not in ['SUCCEEDED', 'FAILED']:
            status = self.get_status(run_id)
            callback(status)

    def get_details(self, run_id, state_name):
        """
        Attempt to extrapolate details from get_status() for a given state_name define in the flow
        definition. Note: This is usually only possible when a flow completes.

        :param run_id: The run_id for this flow. Flow id is automatically determined based
                          on the current tool being run.
        :param state_name: The state in the automate definition to fetch
        :returns: sub-dict of get_status() describing the :state_name:.
        """
        return gladier.utils.automate.get_details(self.get_status(run_id), state_name)



import os
from setuptools import find_packages, setup

# Single source of truth for version
version_ns = {}
with open(os.path.join("gladier", "version.py")) as f:
    exec(f.read(), version_ns)
version = version_ns['__version__']

with open('README.rst') as f:
    long_description = f.read()

install_requires = []
with open('requirements.txt') as reqs:
    for line in reqs.readlines():
        req = line.strip()
        if not req or req.startswith('#'):
            continue
        install_requires.append(req)

setup(
    name="gladier",
    description="Tooling for rapid deployment of automation tooling.",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/globus-gladier/gladier',
    version=version,
    packages=find_packages(),
    install_requires=install_requires,
    python_requires=">=3.6",
    license='Apache 2.0',
    maintainer='',
    maintainer_email='',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ]
)



from .login_manager import BaseLoginManager, AutoLoginManager, CallbackLoginManager
from .flows_manager import FlowsManager
from .funcx_manager import FuncXManager

__all__ = [
    "BaseLoginManager", "AutoLoginManager", "CallbackLoginManager",

    "FlowsManager",

    "FuncXManager",
]



import copy
from typing import List, Mapping, Any
from gladier.utils.tool_alias import ToolAlias
from gladier.utils.flow_traversal import iter_flow


class GladierBaseTool(object):
    """Gladier Defaults defines a common method of tying together
    flows, funcx-functions, and default inputs for starting a flow."""

    flow_definition = None
    flow_input = dict()
    flow_transition_states = []
    required_input = []
    alias_exempt = ['funcx_endpoint_compute', 'funcx_endpoint_non_compute']
    funcx_endpoints = dict()
    funcx_functions = []

    def __init__(self, alias: str = None, alias_class: ToolAlias = None):
        self.alias = alias
        alias_cls = alias_class
        if alias and not alias_class:
            raise ValueError(
                f'{self.__class__.__name__} given alias "{alias}" but not "alias_class". '
                'ex: alias_class=gladier.utils.tool_alias.StateSuffixVariablePrefix'
            )
        if alias_class:
            self.alias_renamer = alias_cls(alias)

    def get_required_input(self) -> List[str]:
        if self.alias:
            required = []
            for input_var in self.required_input:
                if input_var not in self.alias_exempt:
                    required.append(self.alias_renamer.rename_variable(input_var, self))
                else:
                    required.append(input_var)
            return required
        else:
            return self.required_input

    def get_flow_input(self) -> Mapping[str, Any]:
        if not self.alias:
            return self.flow_input
        flow_input = dict()
        for input_var, val in self.flow_input.items():
            if input_var not in self.alias_exempt:
                flow_input[self.alias_renamer.rename_variable(input_var, self)] = val
            else:
                flow_input[input_var] = val
        return flow_input

    def get_flow_transition_states(self) -> List[str]:
        return self.flow_transition_states

    def get_original_inputs(self) -> Mapping[str, Any]:
        return [input_var for input_var in set(self.required_input) | set(self.flow_input.keys())
                if input_var not in self.alias_exempt]

    def rename_state(self, state_name: str, state_data: Mapping[str, Any]) -> Mapping[str, Any]:
        name = self.alias_renamer.rename_state(state_name, self)
        data = self.alias_renamer.rename_input_variables(state_data,
                                                         self.get_original_inputs(), self)
        return name, data

    def get_flow_definition(self) -> Mapping[str, Any]:

        if not self.alias:
            return self.flow_definition

        new_states = {}
        for state_name, state_data in iter_flow(self.flow_definition):
            new_name, new_state = self.rename_state(state_name, state_data)
            new_states[new_name] = new_state
            if new_state.get('Next'):
                new_state['Next'] = self.alias_renamer.rename_state(new_state['Next'], self)
            if new_state['Type'] == 'Choice':
                for choice in new_state['Choices']:
                    if choice.get('Next'):
                        choice['Next'] = self.alias_renamer.rename_state(choice['Next'], self)

        new_flow_def = copy.deepcopy(self.flow_definition)
        new_flow_def['States'] = new_states
        new_flow_def['StartAt'] = self.alias_renamer.rename_state(self.flow_definition['StartAt'],
                                                                  self)
        return new_flow_def



import re


def get_upper_camel_case(snake_case_name):
    """Make a snake_case_name into UpperCamelCase"""
    return ''.join([nb.capitalize() for nb in snake_case_name.split('_')])


def get_snake_case(upper_camel_case):
    # https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    snake_name = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', upper_camel_case)
    snake_name = re.sub('([a-z0-9])([A-Z])', r'\1_\2', snake_name).lower()
    return snake_name


def get_funcx_flow_state_name(funcx_function):
    """State names in Automate are typically upper camel case. This function generates
    an upper case funcx function name for flow states."""
    return get_upper_camel_case(funcx_function.__name__)


def get_funcx_function_name(funcx_function):
    """
    Generate a function name given a funcx function. These function names are used to refer
    to funcx functions within the config. There is no guarantee of uniqueness for function
    names.

    :return: human readable string identifier for a function (intended for a gladier.cfg file)
    """
    return f'{funcx_function.__name__}_funcx_id'


def get_funcx_function_checksum_name(funcx_function):
    """
    Generate a name to refer to the checksum for a given funcx function. Based off of the
    name generated for the function self.get_funcx_function_name. Human readable, intended
    for config.

    :return:  human readable string identifier for a function checksum (for a gladier.cfg file)
    """
    return f'{get_funcx_function_name(funcx_function)}_checksum'
